# Copyright (C) Raffaele Salmaso <raffaele@salmaso.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.

import os
import unittest

import django

import django_service_urls.loads  # noqa: F401

os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"
django.setup()


class MonkeyPatchDjangoTestCase(unittest.TestCase):
    def test_databases(self) -> None:
        from django.conf import settings

        DATABASES = settings.DATABASES
        default_database = DATABASES["default"]
        self.assertTrue(isinstance(default_database, dict))
        self.assertEqual(default_database["ENGINE"], "django.db.backends.postgresql")
        self.assertEqual(default_database["NAME"], "mydb")
        analytics_database = DATABASES["analytics"]
        self.assertTrue(isinstance(analytics_database, dict))
        self.assertEqual(analytics_database["ENGINE"], "django.db.backends.sqlite3")
        self.assertEqual(analytics_database["NAME"], "/tmp/analytics.db")

    def test_caches(self) -> None:
        from django.conf import settings

        CACHES = settings.CACHES
        default_cache = CACHES["default"]
        self.assertTrue(isinstance(default_cache, dict))
        self.assertEqual(default_cache["BACKEND"], "django.core.cache.backends.locmem.LocMemCache")

    def test_email(self) -> None:
        from django.conf import settings

        self.assertEqual(settings.EMAIL_BACKEND, "django.core.mail.backends.smtp.EmailBackend")
        self.assertEqual(settings.EMAIL_HOST, "smtpserver")
        self.assertEqual(settings.EMAIL_PORT, 42)
        self.assertEqual(settings.EMAIL_HOST_USER, "myuser")
        self.assertEqual(settings.EMAIL_HOST_PASSWORD, "mypasswd")
        self.assertEqual(settings.EMAIL_USE_TLS, True)
        self.assertEqual(settings.EMAIL_USE_SSL, False)
        self.assertEqual(settings.EMAIL_SSL_CERTFILE, "mycert")
        self.assertEqual(settings.EMAIL_SSL_KEYFILE, None)
        self.assertEqual(settings.EMAIL_TIMEOUT, 30)

    def test_ensure_that_locale_keep_the_right_path(self) -> None:
        import datetime as dt

        from django.template import engines
        from django.utils import translation

        engine = engines.all()[0]
        template = engine.from_string("{{ now }}")

        with translation.override("en"):
            output = template.render(context={"now": dt.date(2025, 3, 21)})
            self.assertEqual(output, "March 21, 2025")
        with translation.override("it"):
            output = template.render(context={"now": dt.date(2025, 3, 21)})
            self.assertEqual(output, "21 Marzo 2025")



# Copyright (C) Raffaele Salmaso <raffaele@salmaso.org>
# Copyright (C) Tom Forbes
# Copyright (C) Kenneth Reitz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.


def get_version(version):
    """Returns a PEP 386-compliant version number from VERSION."""
    assert len(version) == 5
    assert version[3] in ("alpha", "beta", "rc", "final", "post")

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|c}N - for alpha, beta and rc releases

    parts = 2 if version[2] == 0 else 3
    main = ".".join(str(x) for x in version[:parts])

    sub = ""

    if version[3] != "final":
        mapping = {"alpha": "a", "beta": "b", "rc": "c", "post": ".post"}
        sub = mapping[version[3]] + str(version[4])

    return str(main + sub)



# Copyright (C) Raffaele Salmaso <raffaele@salmaso.org>
# Copyright (C) Tom Forbes
# Copyright (C) Kenneth Reitz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.

import unittest

from django_service_urls import cache, db, email, Service

GENERIC_TESTS = [
    ("username:password@domain/database", ("username", "password", "domain", "", "database", {})),
    ("username:password@domain:123/database", ("username", "password", "domain", 123, "database", {})),
    ("domain:123/database", ("", "", "domain", 123, "database", {})),
    ("user@domain:123/database", ("user", "", "domain", 123, "database", {})),
    (
        "username:password@[2001:db8:1234::1234:5678:90af]:123/database",
        ("username", "password", "2001:db8:1234::1234:5678:90af", 123, "database", {}),
    ),
    (
        "username:password@host:123/database?reconnect=true",
        ("username", "password", "host", 123, "database", {"reconnect": True}),
    ),
    ("username:password@/database", ("username", "password", "", "", "database", {})),
]


class ServiceTestCase(unittest.TestCase):
    def test_an_empty_string_should_returns_an_empty_dict(self) -> None:
        class TestService(Service):
            def config_from_url(self, engine, scheme, url):
                return {}

        service = TestService()
        self.assertEqual(service.parse(""), {})


class DatabaseTestCase(unittest.TestCase):
    SCHEME = None
    STRING_PORTS = False  # Workaround for Oracle

    def test_parsing(self) -> None:
        if self.SCHEME is None:
            return
        for value, (user, passw, host, port, database, options) in GENERIC_TESTS:
            value = f"{self.SCHEME}://{value}"
            with self.subTest(value=value):
                result = db.parse(value)
                self.assertEqual(result["NAME"], database)
                self.assertEqual(result["HOST"], host)
                self.assertEqual(result["USER"], user)
                self.assertEqual(result["PASSWORD"], passw)
                self.assertEqual(result["PORT"], port if not self.STRING_PORTS else str(port))
                self.assertDictEqual(result["OPTIONS"], options)


class SqliteTests(unittest.TestCase):
    SCHEME = "sqlite"

    def test_empty_url(self) -> None:
        url = db.parse("sqlite://")
        self.assertEqual(url["ENGINE"], "django.db.backends.sqlite3")
        self.assertEqual(url["NAME"], ":memory:")

    def test_memory_url(self) -> None:
        url = db.parse("sqlite://:memory:")
        self.assertEqual(url["ENGINE"], "django.db.backends.sqlite3")
        self.assertEqual(url["NAME"], ":memory:")

    def _test_file(self, dbname) -> None:
        url = db.parse("sqlite://{}".format(dbname))
        self.assertEqual(url["ENGINE"], "django.db.backends.sqlite3")
        self.assertEqual(url["NAME"], dbname)
        self.assertEqual(url["HOST"], "")
        self.assertEqual(url["USER"], "")
        self.assertEqual(url["PASSWORD"], "")
        self.assertEqual(url["PORT"], "")
        self.assertDictEqual(url["OPTIONS"], {})

    def test_file(self) -> None:
        self._test_file("/home/user/projects/project/app.sqlite3")
        self._test_file("C:/home/user/projects/project/app.sqlite3")


class PostgresTests(DatabaseTestCase):
    SCHEME = "postgres"

    def test_network_parsing(self) -> None:
        url = db.parse("postgres://uf07k1i6d8ia0v:@:5435/d8r82722r2kuvn")
        self.assertEqual(url["ENGINE"], "django.db.backends.postgresql")
        self.assertEqual(url["NAME"], "d8r82722r2kuvn")
        self.assertEqual(url["HOST"], "")
        self.assertEqual(url["USER"], "uf07k1i6d8ia0v")
        self.assertEqual(url["PASSWORD"], "")
        self.assertEqual(url["PORT"], 5435)

    def test_unix_socket_parsing(self) -> None:
        url = db.parse("postgres://%2Fvar%2Frun%2Fpostgresql/d8r82722r2kuvn")
        self.assertEqual(url["ENGINE"], "django.db.backends.postgresql")
        self.assertEqual(url["NAME"], "d8r82722r2kuvn")
        self.assertEqual(url["HOST"], "/var/run/postgresql")
        self.assertEqual(url["USER"], "")
        self.assertEqual(url["PASSWORD"], "")
        self.assertEqual(url["PORT"], "")

        url = db.parse("postgres://C%3A%2Fvar%2Frun%2Fpostgresql/d8r82722r2kuvn")
        self.assertEqual(url["ENGINE"], "django.db.backends.postgresql")
        self.assertEqual(url["HOST"], "C:/var/run/postgresql")
        self.assertEqual(url["USER"], "")
        self.assertEqual(url["PASSWORD"], "")
        self.assertEqual(url["PORT"], "")

    def test_search_path_schema_parsing(self) -> None:
        url = db.parse(
            "postgres://uf07k1i6d8ia0v:wegauwhgeuioweg@ec2-107-21-253-135.compute-1.amazonaws.com:5431"
            "/d8r82722r2kuvn?currentSchema=otherschema"
        )
        self.assertEqual(url["ENGINE"], "django.db.backends.postgresql")
        self.assertEqual(url["NAME"], "d8r82722r2kuvn")
        self.assertEqual(url["HOST"], "ec2-107-21-253-135.compute-1.amazonaws.com")
        self.assertEqual(url["USER"], "uf07k1i6d8ia0v")
        self.assertEqual(url["PASSWORD"], "wegauwhgeuioweg")
        self.assertEqual(url["PORT"], 5431)
        self.assertEqual(url["OPTIONS"]["options"], "-c search_path=otherschema")
        self.assertNotIn("currentSchema", url["OPTIONS"])

    def test_parsing_with_special_characters(self) -> None:
        url = db.parse("postgres://%23user:%23password@ec2-107-21-253-135.compute-1.amazonaws.com:5431/%23database")
        self.assertEqual(url["ENGINE"], "django.db.backends.postgresql")
        self.assertEqual(url["NAME"], "#database")
        self.assertEqual(url["HOST"], "ec2-107-21-253-135.compute-1.amazonaws.com")
        self.assertEqual(url["USER"], "#user")
        self.assertEqual(url["PASSWORD"], "#password")
        self.assertEqual(url["PORT"], 5431)

    def test_database_url_with_options(self) -> None:
        # Test full options
        url = db.parse(
            "postgres://uf07k1i6d8ia0v:wegauwhgeuioweg"
            "@ec2-107-21-253-135.compute-1.amazonaws.com:5431/d8r82722r2kuvn"
            "?sslrootcert=rds-combined-ca-bundle.pem&sslmode=verify-full"
        )
        self.assertEqual(url["ENGINE"], "django.db.backends.postgresql")
        self.assertEqual(url["NAME"], "d8r82722r2kuvn")
        self.assertEqual(url["HOST"], "ec2-107-21-253-135.compute-1.amazonaws.com")
        self.assertEqual(url["USER"], "uf07k1i6d8ia0v")
        self.assertEqual(url["PASSWORD"], "wegauwhgeuioweg")
        self.assertEqual(url["PORT"], 5431)
        self.assertEqual(url["OPTIONS"], {"sslrootcert": "rds-combined-ca-bundle.pem", "sslmode": "verify-full"})

    def test_gis_search_path_parsing(self) -> None:
        url = db.parse(
            "postgis://uf07k1i6d8ia0v:wegauwhgeuioweg@ec2-107-21-253-135.compute-1.amazonaws.com:5431"
            "/d8r82722r2kuvn?currentSchema=otherschema"
        )
        self.assertEqual(url["ENGINE"], "django.contrib.gis.db.backends.postgis")
        self.assertEqual(url["NAME"], "d8r82722r2kuvn")
        self.assertEqual(url["HOST"], "ec2-107-21-253-135.compute-1.amazonaws.com")
        self.assertEqual(url["USER"], "uf07k1i6d8ia0v")
        self.assertEqual(url["PASSWORD"], "wegauwhgeuioweg")
        self.assertEqual(url["PORT"], 5431)
        self.assertEqual(url["OPTIONS"]["options"], "-c search_path=otherschema")
        self.assertNotIn("currentSchema", url["OPTIONS"])


class MysqlTests(DatabaseTestCase):
    SCHEME = "mysql"

    def test_with_sslca_options(self) -> None:
        url = db.parse(
            "mysql://uf07k1i6d8ia0v:wegauwhgeuioweg"
            "@ec2-107-21-253-135.compute-1.amazonaws.com:3306/d8r82722r2kuvn"
            "?ssl-ca=rds-combined-ca-bundle.pem"
        )
        self.assertEqual(url["ENGINE"], "django.db.backends.mysql")
        self.assertEqual(url["NAME"], "d8r82722r2kuvn")
        self.assertEqual(url["HOST"], "ec2-107-21-253-135.compute-1.amazonaws.com")
        self.assertEqual(url["USER"], "uf07k1i6d8ia0v")
        self.assertEqual(url["PASSWORD"], "wegauwhgeuioweg")
        self.assertEqual(url["PORT"], 3306)
        self.assertEqual(url["OPTIONS"], {"ssl": {"ca": "rds-combined-ca-bundle.pem"}})


class OracleTests(DatabaseTestCase):
    SCHEME = "oracle"
    STRING_PORTS = True

    def test_dsn_parsing(self) -> None:
        dsn = "(DESCRIPTION=(ADDRESS_LIST=(ADDRESS=(PROTOCOL=TCP)(HOST=oraclehost)(PORT=1521)))(CONNECT_DATA=(SID=hr)))"
        url = db.parse("oracle://scott:tiger@/" + dsn)
        self.assertEqual(url["ENGINE"], "django.db.backends.oracle")
        self.assertEqual(url["USER"], "scott")
        self.assertEqual(url["PASSWORD"], "tiger")
        self.assertEqual(url["HOST"], "")
        self.assertEqual(url["PORT"], "")

    def test_empty_dsn_parsing(self) -> None:
        dsn = "(DESCRIPTION=(ADDRESS_LIST=(ADDRESS=(PROTOCOL=TCP)(HOST=oraclehost)(PORT=1521)))(CONNECT_DATA=(SID=hr)))"
        self.assertRaises(ValueError, db.parse, dsn)


class TestCaches(unittest.TestCase):
    def test_local_caching_no_params(self) -> None:
        result = cache.parse("memory://")
        self.assertEqual(result["BACKEND"], "django.core.cache.backends.locmem.LocMemCache")
        self.assertNotIn("LOCATION", result)

    def test_local_caching_with_location(self) -> None:
        result = cache.parse("memory://abc")
        self.assertEqual(result["BACKEND"], "django.core.cache.backends.locmem.LocMemCache")
        self.assertEqual(result["LOCATION"], "abc")

    def test_database_caching(self) -> None:
        result = cache.parse("db://table-name")
        self.assertEqual(result["BACKEND"], "django.core.cache.backends.db.DatabaseCache")
        self.assertEqual(result["LOCATION"], "table-name")

    def test_dummy_caching_no_params(self) -> None:
        result = cache.parse("dummy://")
        self.assertEqual(result["BACKEND"], "django.core.cache.backends.dummy.DummyCache")
        self.assertNotIn("LOCATION", result)

    def test_dummy_caching_with_location(self) -> None:
        result = cache.parse("dummy://abc")
        self.assertEqual(result["BACKEND"], "django.core.cache.backends.dummy.DummyCache")
        self.assertEqual(result["LOCATION"], "abc")

    def test_pymemcached_with_single_ip(self) -> None:
        for url, backend in [
            ("pymemcached://1.2.3.4:1567", "django.core.cache.backends.memcached.PyMemcachedCache"),
            ("memcached://1.2.3.4:1567", "django.core.cache.backends.memcached.MemcachedCache"),
        ]:
            with self.subTest(url=url):
                result = cache.parse(url)
                self.assertEqual(result["BACKEND"], backend)
                self.assertEqual(result["LOCATION"], "1.2.3.4:1567")

    def test_pymemcached_with_multiple_ips(self) -> None:
        for url, backend in [
            ("pymemcached://1.2.3.4:1567,1.2.3.5:1568", "django.core.cache.backends.memcached.PyMemcachedCache"),
            ("memcached://1.2.3.4:1567,1.2.3.5:1568", "django.core.cache.backends.memcached.MemcachedCache"),
        ]:
            with self.subTest(url=url):
                result = cache.parse(url)
                self.assertEqual(result["BACKEND"], backend)
                self.assertEqual(result["LOCATION"], ["1.2.3.4:1567", "1.2.3.5:1568"])

    def test_pymemcached_without_port(self) -> None:
        for url, backend in [
            ("pymemcached://1.2.3.4", "django.core.cache.backends.memcached.PyMemcachedCache"),
            ("memcached://1.2.3.4", "django.core.cache.backends.memcached.MemcachedCache"),
        ]:
            with self.subTest(url=url):
                result = cache.parse(url)
                self.assertEqual(result["BACKEND"], backend)
                self.assertEqual(result["LOCATION"], "1.2.3.4")

    def test_pymemcached_with_unix_socket(self) -> None:
        for url, backend in [
            ("pymemcached:///tmp/memcached.sock", "django.core.cache.backends.memcached.PyMemcachedCache"),
            ("memcached:///tmp/memcached.sock", "django.core.cache.backends.memcached.MemcachedCache"),
        ]:
            with self.subTest(url=url):
                result = cache.parse(url)
                self.assertEqual(result["BACKEND"], backend)
                self.assertEqual(result["LOCATION"], "unix:/tmp/memcached.sock")

    def test_pylibmccache_with_single_ip(self) -> None:
        for url in [
            "pylibmccache://1.2.3.4:1567",
            "memcached+pylibmccache://1.2.3.4:1567",
        ]:
            with self.subTest(url=url):
                result = cache.parse(url)
                self.assertEqual(result["BACKEND"], "django.core.cache.backends.memcached.PyLibMCCache")
                self.assertEqual(result["LOCATION"], "1.2.3.4:1567")

    def test_pylibmccache_with_multiple_ips(self) -> None:
        for url in [
            "pylibmccache://1.2.3.4:1567,1.2.3.5:1568",
            "memcached+pylibmccache://1.2.3.4:1567,1.2.3.5:1568",
        ]:
            with self.subTest(url=url):
                result = cache.parse(url)
                self.assertEqual(result["BACKEND"], "django.core.cache.backends.memcached.PyLibMCCache")
                self.assertEqual(result["LOCATION"], ["1.2.3.4:1567", "1.2.3.5:1568"])

    def test_pylibmccache_without_port(self) -> None:
        for url in [
            "pylibmccache://1.2.3.4",
            "memcached+pylibmccache://1.2.3.4",
        ]:
            with self.subTest(url=url):
                result = cache.parse(url)
                self.assertEqual(result["BACKEND"], "django.core.cache.backends.memcached.PyLibMCCache")
                self.assertEqual(result["LOCATION"], "1.2.3.4")

    def test_pylibmccache_with_unix_socket(self) -> None:
        for url in [
            "pylibmccache:///tmp/memcached.sock",
            "memcached+pylibmccache:///tmp/memcached.sock",
        ]:
            with self.subTest(url=url):
                result = cache.parse(url)
                self.assertEqual(result["BACKEND"], "django.core.cache.backends.memcached.PyLibMCCache")
                self.assertEqual(result["LOCATION"], "/tmp/memcached.sock")

    def test_file_cache_windows_path(self) -> None:
        result = cache.parse("file://C:/abc/def/xyz")
        self.assertEqual(result["BACKEND"], "django.core.cache.backends.filebased.FileBasedCache")
        self.assertEqual(result["LOCATION"], "C:/abc/def/xyz")

    def test_file_cache_unix_path(self) -> None:
        result = cache.parse("file:///abc/def/xyz")
        self.assertEqual(result["BACKEND"], "django.core.cache.backends.filebased.FileBasedCache")
        self.assertEqual(result["LOCATION"], "/abc/def/xyz")


EMAIL_SMTP_TESTS = [
    (
        "smtp://:@:",
        {
            "HOST": "localhost",
            "PORT": 25,
            "HOST_USER": "",
            "HOST_PASSWORD": "",
            "USE_TLS": False,
            "USE_SSL": False,
            "SSL_CERTFILE": None,
            "SSL_KEYFILE": None,
            "TIMEOUT": None,
        },
    ),  # noqa
    (
        "smtps://:@:",
        {
            "HOST": "localhost",
            "PORT": 25,
            "HOST_USER": "",
            "HOST_PASSWORD": "",
            "USE_TLS": True,
            "USE_SSL": False,
            "SSL_CERTFILE": None,
            "SSL_KEYFILE": None,
            "TIMEOUT": None,
        },
    ),  # noqa
    (
        "smtp+tls://:@:",
        {
            "HOST": "localhost",
            "PORT": 25,
            "HOST_USER": "",
            "HOST_PASSWORD": "",
            "USE_TLS": True,
            "USE_SSL": False,
            "SSL_CERTFILE": None,
            "SSL_KEYFILE": None,
            "TIMEOUT": None,
        },
    ),  # noqa
    (
        "smtp+ssl://:@:",
        {
            "HOST": "localhost",
            "PORT": 25,
            "HOST_USER": "",
            "HOST_PASSWORD": "",
            "USE_TLS": False,
            "USE_SSL": True,
            "SSL_CERTFILE": None,
            "SSL_KEYFILE": None,
            "TIMEOUT": None,
        },
    ),  # noqa
]


class EmailsTests(unittest.TestCase):
    def test_smtp(self) -> None:
        for test in EMAIL_SMTP_TESTS:
            url = email.parse(test[0])
            self.assertEqual(url["ENGINE"], "django.core.mail.backends.smtp.EmailBackend")
            for k, v in test[1].items():
                self.assertEqual(url[k], v)

    def test_console(self) -> None:
        url = email.parse("console://")
        self.assertEqual(url["ENGINE"], "django.core.mail.backends.console.EmailBackend")

    def test_file(self) -> None:
        url = email.parse("file://")
        self.assertEqual(url["ENGINE"], "django.core.mail.backends.filebased.EmailBackend")
        self.assertEqual(url["FILE_PATH"], "/")

    def test_memory(self) -> None:
        url = email.parse("memory://")
        self.assertEqual(url["ENGINE"], "django.core.mail.backends.locmem.EmailBackend")

    def test_dummy(self) -> None:
        url = email.parse("dummy://")
        self.assertEqual(url["ENGINE"], "django.core.mail.backends.dummy.EmailBackend")


class TestParseURL(unittest.TestCase):
    def setUp(self) -> None:
        self.backend = Service()

    def test_hostname_sensitivity(self) -> None:
        parsed = self.backend.parse_url("http://CaseSensitive")
        self.assertEqual(parsed["hostname"], "CaseSensitive")

    def test_port_is_an_integer(self) -> None:
        parsed = self.backend.parse_url("http://CaseSensitive:123")
        self.assertIsInstance(parsed["port"], int)

    def test_path_strips_leading_slash(self) -> None:
        parsed = self.backend.parse_url("http://test/abc")
        self.assertEqual(parsed["path"], "abc")

    def test_query_parameters_integer(self) -> None:
        parsed = self.backend.parse_url("http://test/?a=1")
        self.assertDictEqual(parsed["options"], {"a": 1})

    def test_query_parameters_boolean(self) -> None:
        parsed = self.backend.parse_url("http://test/?a=true&b=false")
        self.assertDictEqual(parsed["options"], {"a": True, "b": False})

    def test_query_last_parameter(self) -> None:
        parsed = self.backend.parse_url("http://test/?a=one&a=two")
        self.assertDictEqual(parsed["options"], {"a": "two"})

    def test_does_not_reparse(self) -> None:
        parsed = self.backend.parse_url("http://test/abc")
        self.assertIs(self.backend.parse_url(parsed), parsed)


class DictionaryTests(unittest.TestCase):
    def test_databases(self) -> None:
        result = db.parse(
            {
                "default": "sqlite://:memory:",
                "postgresql": "postgres://uf07k1i6d8ia0v:@:5435/d8r82722r2kuvn",
                "mysql": {
                    "ENGINE": "django.db.backends.mysql",
                    "NAME": "d8r82722r2kuvn",
                    "HOST": "ec2-107-21-253-135.compute-1.amazonaws.com",
                    "USER": "uf07k1i6d8ia0v",
                    "PASSWORD": "wegauwhgeuioweg",
                    "PORT": 3306,
                },
            }
        )
        self.assertEqual(result["default"]["ENGINE"], "django.db.backends.sqlite3")
        self.assertEqual(result["default"]["NAME"], ":memory:")
        self.assertEqual(result["postgresql"]["ENGINE"], "django.db.backends.postgresql")
        self.assertEqual(result["postgresql"]["NAME"], "d8r82722r2kuvn")
        self.assertEqual(result["postgresql"]["HOST"], "")
        self.assertEqual(result["postgresql"]["USER"], "uf07k1i6d8ia0v")
        self.assertEqual(result["postgresql"]["PASSWORD"], "")
        self.assertEqual(result["postgresql"]["PORT"], 5435)
        self.assertEqual(result["mysql"]["ENGINE"], "django.db.backends.mysql")
        self.assertEqual(result["mysql"]["NAME"], "d8r82722r2kuvn")
        self.assertEqual(result["mysql"]["HOST"], "ec2-107-21-253-135.compute-1.amazonaws.com")
        self.assertEqual(result["mysql"]["USER"], "uf07k1i6d8ia0v")
        self.assertEqual(result["mysql"]["PASSWORD"], "wegauwhgeuioweg")
        self.assertEqual(result["mysql"]["PORT"], 3306)

    def test_caches(self) -> None:
        result = cache.parse(
            {
                "default": "memory://",
                "dummy": {
                    "BACKEND": "django.core.cache.backends.dummy.DummyCache",
                },
                "memcached": "memcached://1.2.3.4:1567,1.2.3.5:1568",
            }
        )
        self.assertEqual(result["default"]["BACKEND"], "django.core.cache.backends.locmem.LocMemCache")
        self.assertEqual(result["dummy"]["BACKEND"], "django.core.cache.backends.dummy.DummyCache")
        self.assertEqual(result["memcached"]["BACKEND"], "django.core.cache.backends.memcached.MemcachedCache")
        self.assertEqual(result["memcached"]["LOCATION"], ["1.2.3.4:1567", "1.2.3.5:1568"])


if __name__ == "__main__":
    unittest.main()



# Copyright (C) Raffaele Salmaso <raffaele@salmaso.org>
# Copyright (C) Tom Forbes
# Copyright (C) Kenneth Reitz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.

from urllib import parse

from .base import Service


class DatabaseService(Service):
    def config_from_url(self, engine, scheme, url):
        parsed = self.parse_url(url)
        return {
            "ENGINE": engine,
            "NAME": parse.unquote(parsed["path"] or ""),
            "USER": parse.unquote(parsed["username"] or ""),
            "PASSWORD": parse.unquote(parsed["password"] or ""),
            "HOST": parsed["hostname"],
            "PORT": parsed["port"] or "",
            "OPTIONS": parsed["options"],
        }


db = DatabaseService()


@db.register(
    ("sqlite", "django.db.backends.sqlite3"),
    ("spatialite", "django.contrib.gis.db.backends.spatialite"),
)
def sqlite_config_from_url(backend, engine, scheme, url):
    # These special URLs cannot be parsed correctly.
    if url in ("sqlite://:memory:", "sqlite://"):
        return {
            "ENGINE": engine,
            "NAME": ":memory:",
        }

    parsed = backend.parse_url(url)
    path = "/" + parsed["path"]
    # On windows a path like C:/a/b is parsed with C as the hostname
    # and a/b/ as the path. Reconstruct the windows path here.
    if parsed["hostname"]:
        path = f"{parsed['hostname']}:{path}"
        parsed["location"] = parsed["hostname"] = ""
    parsed["path"] = path
    return backend.config_from_url(engine, scheme, parsed)


@db.register(
    ("postgres", "django.db.backends.postgresql"),
    ("postgis", "django.contrib.gis.db.backends.postgis"),
    # dj_database_url compat aliases
    ("postgresql", "django.db.backends.postgresql"),
    ("pgsql", "django.db.backends.postgresql"),
)
def postgresql_config_from_url(backend, engine, scheme, url):
    parsed = backend.parse_url(url)
    host = parsed["hostname"].lower()
    # Handle postgres percent-encoded paths.
    if "%2f" in host or "%3a" in host:
        parsed["hostname"] = parse.unquote(parsed["hostname"])
    config = backend.config_from_url(engine, scheme, parsed)
    if "currentSchema" in config["OPTIONS"]:
        value = config["OPTIONS"].pop("currentSchema")
        config["OPTIONS"]["options"] = f"-c search_path={value}"
    return config


@db.register(
    ("mysql", "django.db.backends.mysql"),
    ("mysql+gis", "django.contrib.gis.db.backends.mysql"),
)
def mysql_config_from_url(backend, engine, scheme, url):
    config = backend.config_from_url(engine, scheme, url)
    if "ssl-ca" in config["OPTIONS"]:
        value = config["OPTIONS"].pop("ssl-ca")
        config["OPTIONS"]["ssl"] = {"ca": value}
    return config


@db.register(
    ("oracle", "django.db.backends.oracle"),
    ("oracle+gis", "django.contrib.gis.db.backends.oracle"),
)
def oracle_config_from_url(backend, engine, scheme, url):
    config = backend.config_from_url(engine, scheme, url)
    # Oracle requires string ports
    config["PORT"] = str(config["PORT"])
    return config


class CacheService(Service):
    def config_from_url(self, engine, scheme, url, *, multiple_netloc=True):
        parsed = self.parse_url(url, multiple_netloc=multiple_netloc)
        config = {
            "BACKEND": engine,
        }
        if multiple_netloc and parsed["location"]:
            config["LOCATION"] = parsed["location"]
        else:
            if parsed["hostname"]:
                config["LOCATION"] = parsed["hostname"]
                if parsed["port"]:
                    config["LOCATION"] = f"{config['LOCATION']}:{parsed['port']}"
        for key in ("timeout", "key_prefix", "version"):
            if key in parsed["options"]:
                option = parsed["options"].pop(key)
                config[key.upper()] = option
        config["OPTIONS"] = parsed["options"]
        return config


cache = CacheService()


@cache.register(
    ("memory", "django.core.cache.backends.locmem.LocMemCache"),
)
def memory_config_from_url(backend, engine, scheme, url):
    return backend.config_from_url(engine, scheme, url)


@cache.register(
    ("db", "django.core.cache.backends.db.DatabaseCache"),
)
def db_config_from_url(backend, engine, scheme, url):
    return backend.config_from_url(engine, scheme, url)


@cache.register(
    ("dummy", "django.core.cache.backends.dummy.DummyCache"),
)
def dummy_config_from_url(backend, engine, scheme, url):
    return backend.config_from_url(engine, scheme, url)


@cache.register(
    ("pymemcached", "django.core.cache.backends.memcached.PyMemcachedCache"),
    ("memcached", "django.core.cache.backends.memcached.MemcachedCache"),  # for django <= 4.2
)
def pymemcached_config_from_url(backend, engine, scheme, url):
    parsed = backend.parse_url(url, multiple_netloc=True)
    config = backend.config_from_url(engine, scheme, parsed, multiple_netloc=True)
    if parsed["path"]:
        # We are dealing with a URI like pymemcached:///socket/path
        config["LOCATION"] = f"unix:/{parsed['path']}"
    return config


@cache.register(
    ("pylibmccache", "django.core.cache.backends.memcached.PyLibMCCache"),
    ("memcached+pylibmccache", "django.core.cache.backends.memcached.PyLibMCCache"),  # deprecated protocol
)
def pylibmccache_config_from_url(backend, engine, scheme, url):
    # django >= 5.0 remove the "unix:" prefix from unix location
    # keep a different converter until we support old django versions
    parsed = backend.parse_url(url, multiple_netloc=True)
    config = backend.config_from_url(engine, scheme, parsed, multiple_netloc=True)
    if parsed["path"]:
        # We are dealing with a URI like pylibmccache:///socket/path
        config["LOCATION"] = f"/{parsed['path']}"
    return config


@cache.register(
    ("file", "django.core.cache.backends.filebased.FileBasedCache"),
)
def file_config_from_url(backend, engine, scheme, url):
    parsed = backend.parse_url(url)
    config = backend.config_from_url(engine, scheme, parsed)
    path = f"/{parsed['path']}"
    # On windows a path like C:/a/b is parsed with C as the hostname
    # and a/b/ as the path. Reconstruct the windows path here.
    if parsed["hostname"]:
        path = f"{parsed['hostname']}:{path}"
    config["LOCATION"] = path
    return config


class EmailService(Service):
    def config_from_url(self, engine, scheme, url):
        return {
            "ENGINE": engine,
        }


email = EmailService()


@email.register(
    ("smtp", "django.core.mail.backends.smtp.EmailBackend"),
    ("smtps", "django.core.mail.backends.smtp.EmailBackend"),  # smtp+tls alias
    ("smtp+tls", "django.core.mail.backends.smtp.EmailBackend"),
    ("smtp+ssl", "django.core.mail.backends.smtp.EmailBackend"),
)
def email_smtp_config_url(backend, engine, scheme, url):
    config = backend.config_from_url(engine, scheme, url)
    parsed = backend.parse_url(url)
    return {
        "HOST": parsed["hostname"] or "localhost",
        "PORT": parsed["port"] or 25,
        "HOST_USER": parsed["username"] or "",
        "HOST_PASSWORD": parsed["password"] or "",
        "USE_TLS": parsed["options"].get("use_tls", scheme in ("smtps", "smtp+tls")),
        "USE_SSL": parsed["options"].get("use_ssl", scheme == "smtp+ssl"),
        "SSL_CERTFILE": parsed["options"].get("ssl_certfile", None),
        "SSL_KEYFILE": parsed["options"].get("ssl_keyfile", None),
        "TIMEOUT": parsed["options"].get("timeout", None),
        "USE_LOCALTIME": parsed["options"].get("use_localtime", False),
        **config,
    }


@email.register(
    ("console", "django.core.mail.backends.console.EmailBackend"),
)
def email_console_config_url(backend, engine, scheme, url):
    return backend.config_from_url(engine, scheme, url)


@email.register(
    ("file", "django.core.mail.backends.filebased.EmailBackend"),
)
def email_file_config_url(backend, engine, scheme, url):
    config = backend.config_from_url(engine, scheme, url)
    parsed = backend.parse_url(url)
    path = f"/{parsed['path']}"
    # On windows a path like C:/a/b is parsed with C as the hostname
    # and a/b/ as the path. Reconstruct the windows path here.
    if parsed["hostname"]:
        path = f"{parsed['hostname']}:{path}"
    return {
        "FILE_PATH": path,
        **config,
    }


@email.register(
    ("memory", "django.core.mail.backends.locmem.EmailBackend"),
)
def email_memory_config_url(backend, engine, scheme, url):
    return backend.config_from_url(engine, scheme, url)


@email.register(
    ("dummy", "django.core.mail.backends.dummy.EmailBackend"),
)
def email_dummy_config_url(backend, engine, scheme, url):
    return backend.config_from_url(engine, scheme, url)



# Copyright (C) Raffaele Salmaso <raffaele@salmaso.org>
# Copyright (C) Tom Forbes
# Copyright (C) Kenneth Reitz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.

import warnings

import django_service_urls.loads  # noqa: F401

from .deprecation import RemovedInVersion20Warning

warnings.warn(
    "The service_urls.patch module is deprecated in favor of django_service_urls.loads "
    "and will be removed in version 2.0.",
    RemovedInVersion20Warning,
    stacklevel=2,
)



# Copyright (C) Raffaele Salmaso <raffaele@salmaso.org>
# Copyright (C) Tom Forbes
# Copyright (C) Kenneth Reitz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.

import re
from urllib import parse


class Service:
    validation = re.compile(r"^(?P<scheme>\S+)://\S*")

    def config_from_url(self, engine, scheme, url):
        raise NotImplementedError("")

    def __init__(self):
        self._schemes = {}

    def validate(self, data):
        match = self.validation.match(data)
        return match.groups()[0] if match else None

    def _parse(self, data):
        if not isinstance(data, str):
            return data

        if not data:
            return {}

        scheme = self.validate(data)
        if scheme is None:
            raise ValueError(f"{data} is invalid, only full dsn urls (scheme://host...) allowed")
        try:
            _scheme = self._schemes[scheme]
        except KeyError:
            raise ValueError(f"{scheme}:// scheme not registered")
        callback, engine = _scheme["callback"], _scheme["engine"]
        return callback(self, engine, scheme, data)

    def parse(self, data):
        if isinstance(data, dict):
            return {k: self._parse(v) for k, v in data.items()}
        return self._parse(data)

    @staticmethod
    def parse_url(url, *, multiple_netloc=False):
        """
        A method to parse URLs into components that handles quirks
        with the stdlib urlparse, such as lower-cased hostnames.
        Also parses querystrings into typed components.
        """
        # This method may be called with an already parsed URL
        if isinstance(url, dict):
            return url

        # scheme://netloc/path;parameters?query#fragment
        parsed = parse.urlparse(url)
        # 1) cannot have multiple files, so assume that they are always hostnames
        # 2) parsed.hostname always returns a lower-cased hostname
        #    this isn't correct if hostname is a file path, so use '_hostinfo'
        #    to get the actual host
        netlocs = parsed.netloc.split(",") if multiple_netloc else []
        hostname, port = (None, None) if len(netlocs) > 1 else parsed._hostinfo
        if port:
            port = int(port)

        query = parse.parse_qs(parsed.query)
        options = {}
        for key, values in query.items():
            value = values[-1]
            if value.isdigit():
                value = int(value)
            elif value.lower() == "true":
                value = True
            elif value.lower() == "false":
                value = False
            options[key] = value
        path = parsed.path[1:]

        config = {
            "scheme": parsed.scheme,
            "username": parsed.username,
            "password": parsed.password,
            "hostname": hostname,
            "port": port,
            "path": path,
            "fullpath": parsed.path,
            "options": options,
            "location": netlocs if len(netlocs) > 1 else parsed.netloc,
        }
        return config

    def register(self, *args):
        def wrapper(func):
            for scheme, engine in args:
                self._schemes[scheme] = {"callback": func, "engine": engine}
            return func

        return wrapper
